// Solutions to the Advent Of Code 2024.
// Copyright (C) 2025  Stefan Müller
//
// This program is free software: you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.

#include <aoc/common/WeightedEdgeGraph.hpp>

#include <limits>
#include <queue>

WeightedEdgeGraph::WeightedEdgeGraph()
    : firstVertexIncidences_{}, vertexEdgeIncidences_{}, edgeWeights_{}
{
}

int WeightedEdgeGraph::addVertex(const int weight)
{
    firstVertexIncidences_.push_back(-1);
    vertexWeights_.push_back(weight);
    return (int)firstVertexIncidences_.size() - 1;
}

void WeightedEdgeGraph::addEdge(const int vertex1, const int vertex2, const int weight)
{
    vertexEdgeIncidences_.emplace_back(vertex2, firstVertexIncidences_[vertex1]);
    firstVertexIncidences_[vertex1] = (int)vertexEdgeIncidences_.size() - 1;
    vertexEdgeIncidences_.emplace_back(vertex1, firstVertexIncidences_[vertex2]);
    firstVertexIncidences_[vertex2] = (int)vertexEdgeIncidences_.size() - 1;
    edgeWeights_.push_back(weight);
}

int WeightedEdgeGraph::getVertexWeight(const int vertex) const
{
    return vertexWeights_[vertex];
}

int WeightedEdgeGraph::getEdgeWeight(const int edge) const
{
    return edgeWeights_[edge];
}

std::vector<int> WeightedEdgeGraph::dijkstra(const int source) const
{
    std::vector<int> distances(firstVertexIncidences_.size(), std::numeric_limits<int>::max());
    auto compare = [&distances](int left, int right) { return distances[left] > distances[right]; };
    std::priority_queue<int, std::vector<int>, decltype(compare)> queue{ compare };

    distances[source] = 0;
    queue.push(source);

    while (!queue.empty())
    {
        int v{ queue.top() };
        queue.pop();

        for (auto neighbor = begin(v); neighbor != end(); ++neighbor)
        {
            int newDistance{ distances[v] + edgeWeights_[neighbor->edge] };
            if (distances[neighbor->vertex] > newDistance)
            {
                distances[neighbor->vertex] = newDistance;
                queue.push(neighbor->vertex);
            }
        }
    }

    return distances;
}

WeightedEdgeGraph::NeighborIterator WeightedEdgeGraph::begin(const int vertex) const
{
    return { *this, vertex };
}

WeightedEdgeGraph::NeighborIterator WeightedEdgeGraph::end() const
{
    return { *this, -1 };
}
